import torch

class GradMultiply(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale):
        ctx.scale = scale
        res = x.new(x)
        return res
    
    @staticmethod
    def backward(ctx, grad):
        return grad * ctx.scale, None